@@ -146,8 +146,7 @@ module TMail
           ret
         end
 
-        alias to_s encoded
-        alias inspect decoded
+        alias to_s decoded
       ^
     end
 
@@ -156,19 +155,6 @@ module TMail
 
   class HeaderField
 
-    def initialize( fname, fbody, strict )
-      @name = fname
-      fbody.strip!
-      @body = fbody
-      @strict = strict
-
-      @illegal = false
-      @parsed = false
-      @parsing = false
-      # @sync = Sync.new
-    end
-
-
     R     = (1 << 0)  # readable
     W     = (1 << 1)  # writable
     NODUP = (1 << 2)  # dup?
@@ -289,6 +275,22 @@ module TMail
     end
 
 
+    def initialize( fname, fbody, strict )
+      @name = fname
+      fbody.strip!
+      @body = fbody
+      @strict = strict
+
+      @illegal = false
+      @parsed = false
+      @parsing = false
+      # @sync = Sync.new
+    end
+
+    def inspect
+      "#<#{type} #@body>"
+    end
+
     def hash
       @name.hash
     end
@@ -533,6 +535,10 @@ module TMail
     attr_writer :route   # internal use only
 
 
+    def inspect
+      "#<#{type} #{address}>"
+    end
+
     def Address.join( arr )
       arr.collect {|i| ::TMail.quote i }.join('.')
     end
