@@ -1,7 +1,7 @@
 #
 # setup.rb
 #
-#   Copyright (c) 2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
+#   Copyright (c) 2000,2001 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
 #
 #   This program is free software.
 #   You can distribute/modify this program under the terms of
@@ -9,7 +9,13 @@
 #
 
 require 'tempfile'
-require 'rbconfig'
+if i = ARGV.index(/\A--rbconfig=/) then
+  file = $'
+  ARGV.delete_at(i)
+  require file
+else
+  require 'rbconfig'
+end
 
 
 class InstallError < StandardError; end
@@ -17,140 +23,62 @@ class InstallError < StandardError; end
 
 class Installer
 
-  Version   = '2.0.0'
-  Copyright = 'Copyright (c) 2000 Minero Aoki'
+  Version   = '2.0.3'
+  Copyright = 'Copyright (c) 2000,2001 Minero Aoki'
 
 
-  CONFIG = ::Config::CONFIG
-
-  RUBY          = CONFIG['ruby_install_name']
-
-  MAJOR_VERSION = CONFIG['MAJOR'].to_i
-  MINOR_VERSION = CONFIG['MINOR'].to_i
-  TEENY_VERSION = CONFIG['TEENY'].to_i
-  VERSION = CONFIG['MAJOR'] + '.' + CONFIG['MINOR']
-
-  BINDIR     = File.join( CONFIG['bindir'] )
-  LIBDIR     = File.join( CONFIG['libdir'], 'ruby' )
-  DATADIR    = File.join( CONFIG['datadir'] )
-
-  ARCH       = CONFIG['arch']
-
-  NEW_LIB_PATH = ((MAJOR_VERSION >= 2) or
-                  ((MAJOR_VERSION == 1) and
-                   ((MINOR_VERSION >= 5) or
-                    ((MINOR_VERSION == 4) and (TEENY_VERSION >= 4)))))
-    
-  if NEW_LIB_PATH then
-    SITELIBDIR = File.join( LIBDIR, 'site_ruby', VERSION )
-  else
-    SITELIBDIR = File.join( LIBDIR, VERSION, 'site_ruby' )
-  end
-
-  STDLIBDIR  = File.join( LIBDIR, VERSION )
-  RBDIR      = File.join( STDLIBDIR )
-  SODIR      = File.join( STDLIBDIR, ARCH )
-
-  SITE_RB    = File.join( SITELIBDIR )
-  SITE_SO    = File.join( SITELIBDIR, ARCH )
-
-  RUBY_PATH  = File.join( BINDIR, RUBY )
-
-  DLEXT      = CONFIG['DLEXT']
-
-
-  #
-  # name => [ default, argument-name, discription ]
-  #
-  OPTIONS = {
-    'bin-dir'   => [ BINDIR,
-                     'path',
-                     'directory for binary' ],
-    'rb-dir'    => [ SITE_RB,
-                     'path',
-                     'directory for ruby script' ],
-    'so-dir'    => [ SITE_SO,
-                     'path',
-                     'directory for ruby extention' ],
-    'data-dir'  => [ DATADIR,
-                     'path',
-                     'directory for data' ],
-    'ruby-path' => [ RUBY_PATH,
-                     'path',
-                     'path to ruby interpreter' ],
-    'make-prog' => [ 'make',
-                     'name',
-                     'make program for ruby extention' ],
-    'with'      => [ '',
-                     'name,name...',
-                     'package name(s) you want to install' ],
-    'without'   => [ '',
-                     'name,name...',
-                     'package name(s) you do not want to install' ]
-  }
-
-  OPTION_ORDER = %w( bin-dir rb-dir so-dir ruby-path make-prog with without )
-
   TASKS = {
-    'config'       => 'set config option',
-    'setup'        => 'compile extention or else',
-    'install'      => 'install packages',
-    'clean'        => "do `make clean' for each extention",
-    'dryrun'       => 'test run',
-    'show'         => 'show current configuration'
+    'config'       => 'save your config configurations',
+    'setup'        => 'compiles extention or else',
+    'install'      => 'installs packages',
+    'clean'        => "does `make clean' for each extention",
+    'dryrun'       => 'does test run',
+    'show'         => 'shows current configuration'
   }
 
   TASK_ORDER = %w( config setup install clean dryrun show )
 
-  TYPES = %w( bin lib ext share )
+  FILETYPES = %w( bin lib ext share )
 
 
   def initialize( argv )
+    argv = argv.dup
+
     @verbose = true
+    @no_harm = false
+
     @config = {}
     @task = nil
     @other_args = []
 
-    configure
-    parsearg argv.dup
+    @task = parsearg( argv )
+    parsearg_TASK @task, argv
+
+    unless @task == 'config' then
+      load_configs
+      check_packdesig
+    end
   end
 
   attr :config
   attr :task
 
 
-  ConfigFile = 'config.save'
-
-  def configure
-    OPTIONS.each do |k,v|
-      @config[ k ] = v[0]
-    end
-    if File.file? ConfigFile then
-      File.foreach( ConfigFile ) do |line|
-        k, v = line.split( '=', 2 )
-        set_config k.strip, v.strip
-      end
-    end
-  end
-
-  def set_config( k, v )
-    if OPTIONS[k][1] == 'path' then
-      @config[k] = File.expand_path(v)
-    else
-      @config[k] = v
-    end
-  end
+  ###
+  ### arg proc.
+  ###
 
   def parsearg( argv )
-    tasks = /\A(?:#{TASKS.keys.join '|'})\z/
+    task_re = /\A(?:#{TASKS.keys.join '|'})\z/
+    task = nil
     arg = argv.shift
 
     case arg
     when /\A\w+\z/
-      unless tasks === arg then
+      unless task_re === arg then
         raise InstallError, "wrong task: #{arg}"
       end
-      @task = arg
+      task = arg
 
     when '-h', '--help'
       print_usage $stdout
@@ -168,31 +96,31 @@ class Installer
       raise InstallError, "unknown global option '#{arg}'"
     end
 
-    unless argv.empty? then
-      mid = "parsearg_#{@task}"
-      unless respond_to? mid then
-        raise InstallError, "#{@task}:  unknown options: #{argv.join ' '}"
-      end
+    task
+  end
+
+
+  def parsearg_TASK( task, argv )
+    mid = "parsearg_#{task}"
+    if respond_to? mid, true then
       __send__ mid, argv
+    else
+      unless argv.empty? then
+        raise InstallError, "#{task}:  unknown options: #{argv.join ' '}"
+      end
     end
-    check_packdesig
   end
 
   def parsearg_config( args )
-    opts = /\A--(#{OPTIONS.keys.join '|'})=/
-
+    @config_args = {}
+    re = /\A--(#{CONFIG_ORDER.join '|'})=/
     args.each do |i|
-      unless m = opts.match(i) then
-        raise InstallError, "config: unknown option #{i}"
-      end
-      name = m[1]
-      ar   = m.post_match.strip
-      set_config name, ar
+      m = re.match(i) or raise InstallError, "config: unknown option #{i}"
+      @config_args[ m[1] ] = m.post_match.strip
     end
   end
 
   def parsearg_install( args )
-    @no_harm = false
     args.each do |i|
       if i == '--no-harm' then
         @no_harm = true
@@ -203,14 +131,10 @@ class Installer
   end
 
   def parsearg_dryrun( args )
-    @dr_args = args
+    @dryrun_args = args
   end
 
 
-  ###
-  ### global options
-  ###
-
   def print_usage( out )
     out.puts
     out.puts 'Usage:'
@@ -226,16 +150,24 @@ class Installer
     fmt = "  %-20s %s\n"
     out.puts
     out.puts 'Global options:'
-    out.printf fmt, '-h,--help', 'print this message'
+    out.printf fmt, '-h,--help',    'print this message'
     out.printf fmt, '-v,--version', 'print version'
-    out.printf fmt, '--copyright', 'print copyright'
+    out.printf fmt, '--copyright',  'print copyright'
 
     out.puts
     out.puts 'Options for config:'
-    OPTION_ORDER.each do |name|
-      dflt, arg, desc = OPTIONS[name]
+    CONFIG_ORDER.each do |name|
+      dflt, arg, desc, dflt2 = CONFIG_OPTS[name]
+      dflt = dflt2 || dflt
       out.printf "  %-20s %s [%s]\n", "--#{name}=#{arg}", desc, dflt
     end
+    out.printf "  %-20s %s [%s]\n",
+        '--rbconfig=path', 'your rbconfig.rb to load', "running ruby's"
+
+    out.puts
+    out.puts 'Options for install:'
+    out.printf "  %-20s %s [%s]\n",
+        '--no-harm', 'only display what to do if given', 'off'
 
     out.puts
     out.puts 'This archive includes:'
@@ -273,19 +205,27 @@ class Installer
     $stderr.printf "%s done.\n", task
   end
 
+
+  ConfigFile = 'config.save'
+
   def do_config
-    File.open( ConfigFile, 'w' ) do |f|
-      @config.each do |k,v|
-        f.printf "%s=%s\n", k, v if v
-      end
+    CONFIG_OPTS.each do |k,v|
+      dflt, vname, desc = v
+      @config[k] = dflt
     end
+
+    @config_args.each do |k,v|
+      setconf k, v
+    end
+
+    save_configs
   end
 
   def do_show
-    OPTION_ORDER.each do |k|
+    CONFIG_ORDER.each do |k|
       v = @config[k]
       if not v or v.empty? then
-        v = '(not specify)'
+        v = '(not specified)'
       end
       printf "%-10s %s\n", k, v
     end
@@ -302,6 +242,7 @@ class Installer
     }
     into_dir( 'ext' ) {
       foreach_package do
+        clean
         extconf
         make
       end
@@ -310,22 +251,25 @@ class Installer
 
   def do_install
     into_dir( 'bin' ) {
-      foreach_package do |dn, targ|
+      foreach_package do |targ, |
         install_bin
       end
     }
     into_dir( 'lib' ) {
-      foreach_package do |dn, targ|
+      foreach_package do |targ, topfile|
         install_rb targ
+        if topfile then
+          create_topfile targ, topfile
+        end
       end
     }
     into_dir( 'ext' ) {
-      foreach_package do |dn, targ|
+      foreach_package do |targ, |
         install_so targ
       end
     }
     into_dir( 'share' ) {
-      foreach_package do |dn, targ|
+      foreach_package do |targ, |
         install_dat targ
       end
     }
@@ -333,14 +277,15 @@ class Installer
 
   def do_clean
     into_dir( 'ext' ) {
-      foreach_package do |dn, targ|
+      foreach_package do
         clean
       end
     }
+    # rmf ConfigFile
   end
   
   def do_dryrun
-    unless File.directory? 'tmp' then
+    unless dir? 'tmp' then
       $stderr.puts 'setup.rb: setting up temporaly environment...'
       @verbose = $DEBUG
       begin
@@ -359,7 +304,7 @@ class Installer
     exec @config['ruby-path'],
          '-I' + File.join('.', 'tmp', 'lib'),
          '-I' + File.join('.', 'tmp', 'ext'),
-         *@dr_args
+         *@dryrun_args
   end
   
 
@@ -367,13 +312,100 @@ class Installer
   ### lib
   ###
 
-  def into_dir( libn )
-    return unless File.directory? libn
-    chdir( libn ) {
-      yield
-    }
+  #
+  # config
+  #
+
+  c = ::Config::CONFIG
+
+  rubyname = c['ruby_install_name']
+  major = c['MAJOR'].to_i
+  minor = c['MINOR'].to_i
+  teeny = c['TEENY'].to_i
+  version = "#{major}.#{minor}"
+
+  arch = c['arch']
+
+  bindir  = File.join( c['bindir'] )
+  rubylib = File.join( c['libdir'], 'ruby' )
+  datadir = File.join( c['datadir'] )
+
+  rubypath = File.join( bindir, rubyname )
+
+  # >=1.4.4 is new path
+  newpath_p = ((major >= 2) or
+               ((major == 1) and
+                ((minor >= 5) or
+                 ((minor == 4) and (teeny >= 4)))))
+  
+  if newpath_p then
+    sitelibdir = File.join( rubylib, 'site_ruby', version )
+  else
+    sitelibdir = File.join( rubylib, version, 'site_ruby' )
+  end
+  stdlibdir = File.join( rubylib, version )
+
+  siterb = sitelibdir
+  siteso = File.join( sitelibdir, arch )
+
+  CONFIG_OPTS = {
+    'bin-dir'   => [ bindir,
+                     'path',
+                     'directory to install commands' ],
+    'rb-dir'    => [ siterb,
+                     'path',
+                     'directory to install ruby scripts' ],
+    'so-dir'    => [ siteso,
+                     'path',
+                     'directory to install ruby extentions' ],
+    'data-dir'  => [ datadir,
+                     'path',
+                     'directory to install data' ],
+    'ruby-path' => [ rubypath,
+                     'path',
+                     'path to ruby interpreter' ],
+    'make-prog' => [ 'make',
+                     'name',
+                     'make program to compile ruby extentions' ],
+    'with'      => [ '',
+                     'name,name...',
+                     'package name(s) you want to install',
+                     'ALL' ],
+    'without'   => [ '',
+                     'name,name...',
+                     'package name(s) you do not want to install' ]
+  }
+
+  CONFIG_ORDER = %w( bin-dir rb-dir so-dir ruby-path make-prog with without )
+
+  def save_configs
+    File.open( ConfigFile, 'w' ) do |f|
+      @config.each do |k,v|
+        f.printf "%s=%s\n", k, v if v
+      end
+    end
   end
 
+  def load_configs
+    File.file? ConfigFile or raise InstallError, 'setup.rb config first'
+    File.foreach( ConfigFile ) do |line|
+      k, v = line.split( '=', 2 )
+      setconf k.strip, v.strip
+    end
+  end
+
+  def setconf( k, v )
+    if CONFIG_OPTS[k][1] == 'path' then
+      @config[k] = File.expand_path(v)
+    else
+      @config[k] = v
+    end
+  end
+
+
+  #
+  # packages
+  #
 
   def check_packdesig
     @with    = extract_dirs( @config['with'] )
@@ -381,7 +413,7 @@ class Installer
 
     packs = packages
     (@with + @without).each do |i|
-      if not packs.include? i and not File.directory? i then
+      if not packs.include? i and not dir? i then
         raise InstallError, "no such package or directory '#{i}'"
       end
     end
@@ -392,7 +424,7 @@ class Installer
     s.split(',').each do |i|
       if /[\*\?]/ === i then
         tmp = Dir.glob(i)
-        tmp.delete_if {|d| not File.directory? d }
+        tmp.delete_if {|d| not dir? d }
         if tmp.empty? then
           tmp.push i   # causes error
         else
@@ -406,21 +438,11 @@ class Installer
     ret
   end
 
-  def foreach_record( fn = 'PATHCONV' )
-    File.foreach( fn ) do |line|
-      line.strip!
-      next if line.empty?
-      a = line.split(/\s+/, 3)
-      a[2] ||= '.'
-      yield a
-    end
-  end
-
   def packages
     ret = []
-    TYPES.each do |type|
+    FILETYPES.each do |type|
       next unless File.exist? type
-      foreach_record( "#{type}/PATHCONV") do |dir, pack, targ|
+      foreach_record( "#{type}/PATHCONV" ) do |dir, pack, |
         ret.push pack
       end
     end
@@ -429,23 +451,21 @@ class Installer
 
   def foreach_package
     path = {}
-    foreach_record do |dir, *rest|
-      path[dir] = rest
+    foreach_record( './PATHCONV' ) do |dir, pack, targ, topfile, |
+      path[dir] = [pack, targ, topfile]
     end
 
     base = File.basename( Dir.getwd )
     Dir.foreach('.') do |dir|
       next if dir[0] == ?.
-      next unless File.directory? dir
+      next unless dir? dir
 
-      unless path[dir] then
-        raise "abs path for package '#{dir}' not exist"
-      end
-      pack, targ = path[dir]
+      path[dir] or raise "abs path for package '#{dir}' not exist"
+      pack, targ, topfile = path[dir]
 
-      if inclpack( pack, "#{base}/#{dir}" ) then
+      if inclpack pack, "#{base}/#{dir}" then
         chdir( dir ) {
-          yield dir, targ
+          yield targ, topfile
         }
       else
         $stderr.puts "setup.rb: skip #{base}/#{dir}(#{pack}) by user option"
@@ -453,6 +473,16 @@ class Installer
     end
   end
 
+  def foreach_record( fname )
+    File.foreach( fname ) do |line|
+      line.strip!
+      next if line.empty?
+      a = line.split(/\s+/)
+      a[2] ||= '.'
+      yield a
+    end
+  end
+
   def inclpack( pack, dname )
     if @with.empty? then
       not @without.include? pack and
@@ -464,6 +494,10 @@ class Installer
   end
 
 
+  #
+  # setup
+  #
+
   def add_rubypath( fn, opt = nil )
     line = "\#!#{@config['ruby-path']}#{opt ? ' ' + opt : ''}"
 
@@ -473,11 +507,11 @@ class Installer
     tmpf = nil
     File.open( fn ) do |f|
       first = f.gets
-      return unless /\A\#\!.*ruby/ === first
+      return unless /\A\#!.*ruby/ === first
 
       tmpf = Tempfile.open( 'amsetup' )
       tmpf.puts line
-      tmpf << first unless /\A\#\!/o === first
+      tmpf << first
       f.each {|i| tmpf << i }
       tmpf.close
     end
@@ -493,6 +527,10 @@ class Installer
   end
 
 
+  #
+  # install
+  #
+
   def install_bin
     install_all isdir(@config['bin-dir']), 0555
   end
@@ -514,9 +552,19 @@ class Installer
     end
   end
 
+  def create_topfile( name, req )
+    d = isdir(@config['rb-dir'])
+    File.open( "#{d}/#{name}.rb", 'w' ) do |f|
+      f.puts "require '#{name}/#{req}'"
+    end
+    File.chmod 0644, "#{d}/#{name}.rb"
+  end
+
 
   def extconf
-    system "#{@config['ruby-path']} extconf.rb"
+    # system "#{@config['ruby-path']} extconf.rb"
+    restore_mkmf_env
+    load 'extconf.rb'
   end
 
   def make
@@ -524,7 +572,7 @@ class Installer
   end
   
   def clean
-    command @config['make-prog'] + ' clean'
+    command @config['make-prog'] + ' clean' if File.file? 'Makefile'
   end
 
   def install_so( dir )
@@ -534,17 +582,15 @@ class Installer
     end
   end
 
+  DLEXT = ::Config::CONFIG['DLEXT']
+
   def find_so( dir = '.' )
     fnames = nil
     Dir.open( dir ) {|d| fnames = d.to_a }
     exp = /\.#{DLEXT}\z/
     arr = fnames.find_all {|fn| exp === fn }
-    unless arr then
-      raise InstallError,
-        'ruby extention not found: try "ruby setup.rb setup"'
-    end
-
-    arr
+    arr or raise InstallError,
+            'no ruby extention exists: have you done "ruby setup.rb setup" ?'
   end
 
   def so_dir?( dn = '.' )
@@ -552,11 +598,15 @@ class Installer
   end
 
 
-  ## fileutils
+  #
+  # file op.
+  #
 
-  def isdir( dn )
-    mkpath dn
-    dn
+  def into_dir( libn )
+    return unless dir? libn
+    chdir( libn ) {
+      yield
+    }
   end
 
   def chdir( dn )
@@ -569,62 +619,78 @@ class Installer
     end
   end
 
+  def isdir( dn )
+    mkpath dn
+    dn
+  end
+
   def mkpath( dname )
     $stderr.puts "mkdir -p #{dname}" if @verbose
     return if @no_harm
 
-    dirs = dname.split('/')
-    if dirs[0].empty? then
-      dirs.shift
-      path = '/'
-    else
-      path = ''
+    # does not check '/'... it's too abnormal case
+    dirs = dname.split(%r_(?=/)_)
+    if /\A[a-z]:\z/i === dirs[0] then
+      disk = dirs.shift
+      dirs[0] = disk + dirs[0]
     end
-    dirs.each do |fname|
-      path << fname << '/'
-      unless File.directory? path then
-        Dir.mkdir path
-      end
+    dirs.each_index do |idx|
+      path = dirs[0..idx].join('')
+      Dir.mkdir path unless dir? path
     end
   end
 
   def rmf( fname )
-    $stderr.puts "rm #{fname}" if @verbose
+    $stderr.puts "rm -f #{fname}" if @verbose
     return if @no_harm
 
-    File.chmod 777, fname
-    File.unlink fname
+    if File.exist? fname or File.symlink? fname then
+      File.chmod 777, fname
+      File.unlink fname
+    end
   end
 
   def rmrf( dn )
-    $stderr.puts "rm -r #{dn}" if @verbose
+    $stderr.puts "rm -rf #{dn}" if @verbose
     return if @no_harm
 
     Dir.chdir dn
     Dir.foreach('.') do |fn|
       next if fn == '.'
       next if fn == '..'
-      if File.directory? fn then
-        rmrf fn
+      if dir? fn then
+        verbose_off {
+          rmrf fn
+        }
       else
-        rmf fn
+        verbose_off {
+          rmf fn
+        }
       end
     end
     Dir.chdir '..'
     Dir.rmdir dn
   end
 
+  def verbose_off
+    save, @verbose = @verbose, false
+    yield
+    @verbose = save
+  end
+
   def install( from, to, mode )
     $stderr.puts "install #{from} #{to}" if @verbose
     return if @no_harm
 
-    if File.directory? to then
+    if dir? to then
       to = to + '/' + File.basename(from)
     end
     str = nil
     File.open( from, 'rb' ) {|f| str = f.read }
     if diff? str, to then
-      rmf to if File.exist? to
+      verbose_off {
+        rmf to if File.exist? to
+      }
       File.open( to, 'wb' ) {|f| f.write str }
       File.chmod mode, to
     end
@@ -638,23 +704,48 @@ class Installer
   end
 
   def command( str )
-    $stderr.puts "system #{str}" if @verbose
+    $stderr.puts str if @verbose
+    system str or raise RuntimeError, "'system #{str}' failed"
+  end
 
-    unless ret = system( str ) then
-      raise RuntimeError, "'system #{str}' failed"
-    end
-    ret
+  def dir?( dname )
+    # for CORRUPTED windows stat()
+    File.directory? (dname[-1,1] == '/' ? dname : dname + '/')
   end
 
 end
 
 
-begin
-  MainInstaller = Installer.new( ARGV )
-  MainInstaller.execute
-rescue
-  raise if $DEBUG
-  $stderr.puts $!
-  $stderr.puts 'try "ruby setup.rb --help" for usage'
-  exit 1
+require 'mkmf'
+GV = %w( $CFLAGS $CPPFLAGS $DLDFLAGS $LDFLAGS $LIBPATH $LOCAL_LIBS $OBJEXT
+         $archdir $config_cache $defs $hdrdir $libdir $libs $local_flags
+         $make $null $objs $sitearchdir $sitelibdir $srcdir $topdir )
+SAVE_ENV = {}
+GV.each do |i|
+  v = eval(i)
+  SAVE_ENV[i] = v ? v.dup : v
+end
+
+def restore_mkmf_env
+  i = nil
+
+  SAVE_ENV.each do |k,v|
+    eval "#{k} = v"
+  end
+  if defined? $configure_args then
+    $configure_args.clear
+  end
+end
+
+
+if $0 == __FILE__ then
+  begin
+    MainInstaller = Installer.new( ARGV )
+    MainInstaller.execute
+  rescue
+    raise if $DEBUG
+    $stderr.puts $!
+    $stderr.puts 'try "ruby setup.rb --help" for usage'
+    exit 1
+  end
 end
